"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresDatabaseAdapter = exports.PostgresSQLDatabaseQueryFactory = exports.PostgresSQLDatabaseQuery = exports.PostgresSQLQueryResolver = exports.PostgresPersistence = exports.PostgresConnectionPool = exports.PostgresConnection = exports.PostgresStatement = void 0;
const sql_1 = require("@deepkit/sql");
const postgres_platform_1 = require("./postgres-platform");
const type_1 = require("@deepkit/type");
const pg_1 = __importDefault(require("pg"));
const core_1 = require("@deepkit/core");
class PostgresStatement extends sql_1.SQLStatement {
    constructor(sql, client) {
        super();
        this.sql = sql;
        this.client = client;
        this.released = false;
    }
    async get(params = []) {
        return core_1.asyncOperation((resolve, reject) => {
            this.client.query(this.sql, params).then((res) => {
                resolve(res.rows[0]);
            }).catch(reject);
        });
    }
    async all(params = []) {
        return core_1.asyncOperation((resolve, reject) => {
            this.client.query(this.sql, params).then((res) => {
                resolve(res.rows);
            }).catch(reject);
        });
    }
    release() {
    }
}
exports.PostgresStatement = PostgresStatement;
class PostgresConnection extends sql_1.SQLConnection {
    constructor(connectionPool, getConnection) {
        super(connectionPool);
        this.getConnection = getConnection;
        this.changes = 0;
        this.lastReturningRows = [];
        this.connection = undefined;
    }
    release() {
        super.release();
        if (this.connection) {
            this.connection.release();
            this.connection = undefined;
        }
    }
    async prepare(sql) {
        if (!this.connection)
            this.connection = await this.getConnection();
        return new PostgresStatement(sql, this.connection);
    }
    async run(sql, params = []) {
        await core_1.asyncOperation(async (resolve, reject) => {
            if (!this.connection)
                this.connection = await this.getConnection();
            this.connection.query(sql, params).then((res) => {
                this.lastReturningRows = res.rows;
                this.changes = res.rowCount;
                resolve(undefined);
            }, reject);
        });
    }
    async getChanges() {
        return this.changes;
    }
}
exports.PostgresConnection = PostgresConnection;
class PostgresConnectionPool extends sql_1.SQLConnectionPool {
    constructor(pool) {
        super();
        this.pool = pool;
    }
    getConnection() {
        this.activeConnections++;
        return new PostgresConnection(this, () => this.pool.connect());
    }
}
exports.PostgresConnectionPool = PostgresConnectionPool;
function typeSafeDefaultValue(property) {
    if (property.type === 'number')
        return 0;
    if (property.type === 'boolean')
        return false;
    if (property.type === 'date')
        return new Date;
    return null;
}
class PostgresPersistence extends sql_1.SQLPersistence {
    constructor(platform, connection) {
        super(platform, connection);
        this.platform = platform;
        this.connection = connection;
        this.placeholderPosition = 1;
    }
    async update(classSchema, changeSets) {
        const scopeSerializer = this.platform.serializer.for(classSchema);
        const tableName = this.platform.getTableIdentifier(classSchema);
        const pkName = classSchema.getPrimaryField().name;
        const pkField = this.platform.quoteIdentifier(pkName);
        const values = {};
        const setNames = [];
        const aggregateSelects = {};
        const requiredFields = {};
        const assignReturning = {};
        const setReturning = {};
        for (const changeSet of changeSets) {
            const where = [];
            const pk = scopeSerializer.partialSerialize(changeSet.primaryKey);
            for (const i in pk) {
                if (!pk.hasOwnProperty(i))
                    continue;
                where.push(`${this.platform.quoteIdentifier(i)} = ${this.platform.quoteValue(pk[i])}`);
                requiredFields[i] = 1;
            }
            if (!values[pkName])
                values[pkName] = [];
            values[pkName].push(this.platform.quoteValue(changeSet.primaryKey[pkName]));
            const fieldAddedToValues = {};
            const id = changeSet.primaryKey[pkName];
            if (changeSet.changes.$set) {
                const value = scopeSerializer.partialSerialize(changeSet.changes.$set);
                for (const i in value) {
                    if (!value.hasOwnProperty(i))
                        continue;
                    if (!values[i]) {
                        values[i] = [];
                        setNames.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
                    }
                    requiredFields[i] = 1;
                    fieldAddedToValues[i] = 1;
                    let v = value[i];
                    //special postgres check to avoid an error like:
                    /// column "deletedAt" is of type timestamp without time zone but expression is of type text
                    if (v === undefined || v === null) {
                        if (classSchema.getProperty(i).type === 'date') {
                            values[i].push('null::timestamp');
                        }
                        else {
                            values[i].push('null');
                        }
                    }
                    else {
                        values[i].push(this.platform.quoteValue(v));
                    }
                }
            }
            if (changeSet.changes.$inc) {
                for (const i in changeSet.changes.$inc) {
                    if (!changeSet.changes.$inc.hasOwnProperty(i))
                        continue;
                    const value = changeSet.changes.$inc[i];
                    if (!aggregateSelects[i])
                        aggregateSelects[i] = [];
                    if (!values[i]) {
                        values[i] = [];
                        setNames.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
                    }
                    if (!assignReturning[id]) {
                        assignReturning[id] = { item: changeSet.item, names: [] };
                    }
                    assignReturning[id].names.push(i);
                    setReturning[i] = 1;
                    aggregateSelects[i].push({ id: changeSet.primaryKey[pkName], sql: `_origin.${this.platform.quoteIdentifier(i)} + ${this.platform.quoteValue(value)}` });
                    requiredFields[i] = 1;
                    if (!fieldAddedToValues[i]) {
                        fieldAddedToValues[i] = 1;
                        values[i].push(this.platform.quoteValue(typeSafeDefaultValue(classSchema.getProperty(i))));
                    }
                }
            }
        }
        const selects = [];
        const valuesValues = [];
        const valuesNames = [];
        for (const i in values) {
            valuesNames.push(i);
        }
        for (let i = 0; i < values[pkName].length; i++) {
            valuesValues.push('(' + valuesNames.map(name => values[name][i]).join(',') + ')');
        }
        for (const i in requiredFields) {
            if (aggregateSelects[i]) {
                const select = [];
                select.push('CASE');
                for (const item of aggregateSelects[i]) {
                    select.push(`WHEN _.${pkField} = ${item.id} THEN ${item.sql}`);
                }
                select.push(`ELSE _.${this.platform.quoteIdentifier(i)} END as ${this.platform.quoteIdentifier(i)}`);
                selects.push(select.join(' '));
            }
            else {
                selects.push('_.' + this.platform.quoteIdentifier(i));
            }
        }
        const returningSelect = [];
        returningSelect.push(tableName + '.' + pkField);
        if (!core_1.empty(setReturning)) {
            for (const i in setReturning) {
                returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(i));
            }
        }
        const escapedValuesNames = valuesNames.map(v => this.platform.quoteIdentifier(v));
        const sql = `
              WITH _b(${escapedValuesNames.join(', ')}) AS (
                SELECT ${selects.join(', ')} FROM 
                    (VALUES ${valuesValues.join(', ')}) as _(${escapedValuesNames.join(', ')})
                    INNER JOIN ${tableName} as _origin ON (_origin.${pkField} = _.${pkField})
              )
              UPDATE ${tableName}
              SET ${setNames.join(', ')}
              FROM _b
              WHERE ${tableName}.${pkField} = _b.${pkField}
              RETURNING ${returningSelect.join(', ')};
        `;
        const result = await this.connection.execAndReturnAll(sql);
        for (const returning of result) {
            const r = assignReturning[returning[pkName]];
            if (!r)
                continue;
            for (const name of r.names) {
                r.item[name] = returning[name];
            }
        }
    }
    async populateAutoIncrementFields(classSchema, items) {
        const autoIncrement = classSchema.getAutoIncrementField();
        if (!autoIncrement)
            return;
        //We adjusted the INSERT SQL with additional RETURNING which returns all generated
        //auto-increment values. We read the result and simply assign the value.
        const name = autoIncrement.name;
        const insertedRows = this.connection.lastReturningRows;
        if (!insertedRows.length)
            return;
        for (let i = 0; i < items.length; i++) {
            items[i][name] = insertedRows[i][name];
        }
    }
    getInsertSQL(classSchema, fields, values) {
        const autoIncrement = classSchema.getAutoIncrementField();
        const returning = autoIncrement ? ` RETURNING ${this.platform.quoteIdentifier(autoIncrement.name)}` : '';
        return `INSERT INTO ${this.platform.getTableIdentifier(classSchema)} (${fields.join(', ')}) VALUES (${values.join('), (')}) ${returning}`;
    }
    resetPlaceholderSymbol() {
        this.placeholderPosition = 1;
    }
    getPlaceholderSymbol() {
        return '$' + this.placeholderPosition++;
    }
}
exports.PostgresPersistence = PostgresPersistence;
class PostgresSQLQueryResolver extends sql_1.SQLQueryResolver {
    async delete(model, deleteResult) {
        const primaryKey = this.classSchema.getPrimaryField();
        const pkField = this.platform.quoteIdentifier(primaryKey.name);
        const primaryKeyConverted = type_1.getPropertyXtoClassFunction(primaryKey, this.platform.serializer);
        const sqlBuilder = new sql_1.SqlBuilder(this.platform);
        const select = sqlBuilder.select(this.classSchema, model, { select: [pkField] });
        const tableName = this.platform.getTableIdentifier(this.classSchema);
        const connection = this.connectionPool.getConnection();
        try {
            const sql = `
                WITH _ AS (${select.sql})
                DELETE
                FROM ${tableName} USING _
                WHERE ${tableName}.${pkField} = _.${pkField}
                RETURNING ${tableName}.${pkField}
            `;
            const rows = await connection.execAndReturnAll(sql, select.params);
            deleteResult.modified = rows.length;
            for (const row of rows) {
                deleteResult.primaryKeys.push(primaryKeyConverted(row[primaryKey.name]));
            }
        }
        finally {
            connection.release();
        }
    }
    async patch(model, changes, patchResult) {
        const select = [];
        const tableName = this.platform.getTableIdentifier(this.classSchema);
        const primaryKey = this.classSchema.getPrimaryField();
        const pkField = this.platform.quoteIdentifier(primaryKey.name);
        const primaryKeyConverted = type_1.getPropertyXtoClassFunction(primaryKey, this.platform.serializer);
        select.push(pkField);
        const fieldsSet = {};
        const aggregateFields = {};
        const scopeSerializer = this.platform.serializer.for(this.classSchema);
        const $set = changes.$set ? scopeSerializer.partialSerialize(changes.$set) : undefined;
        const set = [];
        if ($set)
            for (const i in $set) {
                if (!$set.hasOwnProperty(i))
                    continue;
                if ($set[i] === undefined || $set[i] === null) {
                    set.push(`${this.platform.quoteIdentifier(i)} = NULL`);
                }
                else {
                    fieldsSet[i] = 1;
                    select.push(`${this.platform.quoteValue($set[i])} as ${this.platform.quoteIdentifier(i)}`);
                }
            }
        if (changes.$unset)
            for (const i in changes.$unset) {
                if (!changes.$unset.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                select.push(`NULL as ${this.platform.quoteIdentifier(i)}`);
            }
        for (const i of model.returning) {
            aggregateFields[i] = { converted: type_1.getPropertyXtoClassFunction(type_1.resolvePropertySchema(this.classSchema, i), this.platform.serializer) };
            select.push(`(${this.platform.quoteIdentifier(i)} ) as ${this.platform.quoteIdentifier(i)}`);
        }
        if (changes.$inc)
            for (const i in changes.$inc) {
                if (!changes.$inc.hasOwnProperty(i))
                    continue;
                fieldsSet[i] = 1;
                aggregateFields[i] = { converted: type_1.getPropertyXtoClassFunction(type_1.resolvePropertySchema(this.classSchema, i), this.platform.serializer) };
                select.push(`(${this.platform.quoteIdentifier(i)} + ${this.platform.quoteValue(changes.$inc[i])}) as ${this.platform.quoteIdentifier(i)}`);
            }
        for (const i in fieldsSet) {
            set.push(`${this.platform.quoteIdentifier(i)} = _b.${this.platform.quoteIdentifier(i)}`);
        }
        const returningSelect = [];
        returningSelect.push(tableName + '.' + pkField);
        if (!core_1.empty(aggregateFields)) {
            for (const i in aggregateFields) {
                returningSelect.push(tableName + '.' + this.platform.quoteIdentifier(i));
            }
        }
        const sqlBuilder = new sql_1.SqlBuilder(this.platform);
        const selectSQL = sqlBuilder.select(this.classSchema, model, { select });
        const sql = `
            WITH _b AS (${selectSQL.sql})
            UPDATE
                ${tableName}
            SET
                ${set.join(', ')}
            FROM _b
            WHERE ${tableName}.${pkField} = _b.${pkField}
            RETURNING ${returningSelect.join(', ')}
        `;
        const connection = this.connectionPool.getConnection();
        try {
            const result = await connection.execAndReturnAll(sql, selectSQL.params);
            patchResult.modified = result.length;
            for (const i in aggregateFields) {
                patchResult.returning[i] = [];
            }
            for (const returning of result) {
                patchResult.primaryKeys.push(primaryKeyConverted(returning[primaryKey.name]));
                for (const i in aggregateFields) {
                    patchResult.returning[i].push(aggregateFields[i].converted(returning[i]));
                }
            }
        }
        finally {
            connection.release();
        }
    }
}
exports.PostgresSQLQueryResolver = PostgresSQLQueryResolver;
class PostgresSQLDatabaseQuery extends sql_1.SQLDatabaseQuery {
}
exports.PostgresSQLDatabaseQuery = PostgresSQLDatabaseQuery;
class PostgresSQLDatabaseQueryFactory extends sql_1.SQLDatabaseQueryFactory {
    createQuery(classType) {
        return new PostgresSQLDatabaseQuery(type_1.getClassSchema(classType), this.databaseSession, new PostgresSQLQueryResolver(this.connectionPool, this.platform, type_1.getClassSchema(classType), this.databaseSession));
    }
}
exports.PostgresSQLDatabaseQueryFactory = PostgresSQLDatabaseQueryFactory;
class PostgresDatabaseAdapter extends sql_1.SQLDatabaseAdapter {
    constructor(options) {
        super();
        this.options = options;
        this.pool = new pg_1.default.Pool(this.options);
        this.connectionPool = new PostgresConnectionPool(this.pool);
        this.platform = new postgres_platform_1.PostgresPlatform();
        pg_1.default.types.setTypeParser(1700, parseFloat);
        pg_1.default.types.setTypeParser(20, BigInt);
    }
    getName() {
        return 'postgres';
    }
    getSchemaName() {
        //todo extract schema name from connection options. This acts as default when a table has no schemaName defined.
        return '';
    }
    createPersistence() {
        return new PostgresPersistence(this.platform, this.connectionPool.getConnection());
    }
    queryFactory(databaseSession) {
        return new PostgresSQLDatabaseQueryFactory(this.connectionPool, this.platform, databaseSession);
    }
    disconnect(force) {
        this.pool.end().catch(console.error);
    }
}
exports.PostgresDatabaseAdapter = PostgresDatabaseAdapter;
//# sourceMappingURL=postgres-adapter.js.map